Macro set_build_debug
set_build_debug!() { /* proc-macro */ }Expand description
Same as set_build, but prints the generated set builder to stderr
set_build_debug!() { /* proc-macro */ }Same as set_build, but prints the generated set builder to stderr